import numpy as np
from numpy.random import multivariate_normal, rand, normal
from numpy.linalg import solve, det
from scipy.integrate import trapezoid, cumulative_trapezoid
from scipy.stats import truncnorm, norm
import fdasrsf.utility_functions as uf
import cbayesian as bay


def f_warp_pw(v, q1, q2):
    obs_domain = np.linspace(0, 1, v.shape[0])
    exp1g_temp = uf.f_predictfunction(uf.f_exp1(v), obs_domain, 0)
    pt = np.insert(bay.bcuL2norm2(obs_domain, exp1g_temp), 0, 0)
    tmp = uf.f_predictfunction(q2, pt, 0)
    out = tmp * exp1g_temp

    return out


def f_SSEv_pw(v, q1, q2):
    q2_gamma = f_warp_pw(v, q1, q2)
    vec = (q1 - q2_gamma) ** 2
    out = vec.sum()

    return out


def f_vpostlogl_pw(v, q1, q2, var, SSEv):
    if SSEv == 0:
        SSEv = f_SSEv_pw(v, q1, q2)

    n = q1.shape[0]
    out = -n * np.log(np.sqrt(var)) - SSEv / (2 * var)
    return out, SSEv


def f_f1postlogl_pw(f1, y1, SSEv, K_f1, sigma_curr, sigma1_curr):
    n = y1.shape[0]
    iSig_f1 = K_f1 + np.eye(n) * n / sigma1_curr
    f1_mean = solve(iSig_f1, y1) * n / sigma1_curr
    SSEf1 = (f1 - f1_mean) @ iSig_f1 @ (f1 - f1_mean)
    out = -SSEv / (2 * sigma_curr) - SSEf1 / 2
    return out


def f_f2postlogl_pw(f2, y2, SSEv, K_f2, sigma_curr, sigma2_curr):
    n = y2.shape[0]
    iSig_f2 = K_f2 + np.eye(n) * n / sigma2_curr
    f2_mean = solve(iSig_f2, y2) * n / sigma2_curr
    SSEf2 = (f2 - f2_mean) @ iSig_f2 @ (f2 - f2_mean)
    out = -SSEv / (2 * sigma_curr) - SSEf2 / 2
    return out


def f_dlogl_pw(v_coef, v_basis, d_basis, sigma_curr, q1, q2):
    vec = uf.f_basistofunction(v_basis["x"], 0, v_coef, v_basis)
    psi = uf.f_exp1(vec)
    N = q1.shape[0]
    obs_domain = np.linspace(0, 1, N)
    binsize = np.diff(obs_domain)
    binsize = binsize.mean()
    gamma = uf.f_phiinv(psi)
    q2_warp = uf.warp_q_gamma(obs_domain, q2, gamma)
    q2_warp_grad = np.gradient(q2_warp, binsize)

    basismat = d_basis["matrix"]

    g = np.zeros(N)
    for i in range(0, basismat.shape[1]):
        ubar = cumulative_trapezoid(basismat[:, i], obs_domain, initial=0)
        integrand = (q1 - q2_warp) * (
            -2 * q2_warp_grad * ubar - q2_warp * basismat[:, i]
        )
        tmp = 1 / sigma_curr * trapezoid(integrand, obs_domain)
        g += tmp * basismat[:, i]

    out, SSEv = f_vpostlogl_pw(vec, q1, q2, sigma_curr, 0)

    nll = -1 * out
    g_coef = v_basis["matrix"].T @ g

    return nll, g_coef, SSEv


def f_updatef1_pw(
    f1_curr,
    q1_curr,
    y1,
    q2,
    v_coef_curr,
    v_basis,
    SSE_curr,
    K_f1,
    K_f1prop,
    sigma_curr,
    sigma1_curr,
):
    time = np.linspace(0, 1, y1.shape[0])
    v = uf.f_basistofunction(v_basis["x"], 0, v_coef_curr, v_basis)

    f1_prop = multivariate_normal(f1_curr, K_f1prop)
    q1_prop = uf.f_to_srsf(f1_prop, time)

    SSE_prop = f_SSEv_pw(v, q1_prop, q2)

    postlog_curr = f_f1postlogl_pw(f1_curr, y1, SSE_curr, K_f1, sigma_curr, sigma1_curr)
    postlog_prop = f_f1postlogl_pw(f1_prop, y1, SSE_prop, K_f1, sigma_curr, sigma1_curr)

    ratio = np.minimum(1, np.exp(postlog_prop - postlog_curr))

    u = rand()
    if u <= ratio:
        f1_curr = f1_prop
        q1_curr = q1_prop
        f1_accept = True
    else:
        f1_accept = False

    return f1_curr, q1_curr, f1_accept


def f_updatef2_pw(
    f2_curr,
    q2_curr,
    y2,
    q1,
    v_coef_curr,
    v_basis,
    SSE_curr,
    K_f2,
    K_f2prop,
    sigma_curr,
    sigma2_curr,
):
    time = np.linspace(0, 1, y2.shape[0])
    v = uf.f_basistofunction(v_basis["x"], 0, v_coef_curr, v_basis)

    f2_prop = multivariate_normal(f2_curr, K_f2prop)
    q2_prop = uf.f_to_srsf(f2_prop, time)

    SSE_prop = f_SSEv_pw(v, q1, q2_prop)

    postlog_curr = f_f2postlogl_pw(f2_curr, y2, SSE_curr, K_f2, sigma_curr, sigma2_curr)
    postlog_prop = f_f2postlogl_pw(f2_prop, y2, SSE_prop, K_f2, sigma_curr, sigma2_curr)

    ratio = np.minimum(1, np.exp(postlog_prop - postlog_curr))

    u = rand()
    if u <= ratio:
        f2_curr = f2_prop
        q2_curr = q2_prop
        f2_accept = True
    else:
        f2_accept = False

    return f2_curr, q2_curr, f2_accept


def f_updatephi_pw(f1_curr, K_f1, s1_curr, L1_curr, L1_propvar, Dmat):
    a, b = (L1_curr - L1_curr) / L1_propvar, (100000 - L1_curr) / L1_propvar
    L1_prop = truncnorm.rvs(a, b)

    K_f1_tmp = s1_curr * (uf.exp2corr2(L1_prop, Dmat) + 0.1 * np.eye(f1_curr.shape[0]))

    SSEf_curr = (f1_curr @ K_f1 @ f1_curr) / 2
    SSEf_prop = (uf.mrdivide(f1_curr, K_f1_tmp) @ f1_curr.T) / 2

    postlog_prop = (
        -np.log(det(K_f1_tmp)) / 2
        - SSEf_prop
        - np.log(1 - norm.cdf(-L1_curr / L1_propvar))
    )
    postlog_curr = (
        np.log(det(K_f1)) / 2 - SSEf_curr - np.log(1 - norm.cdf(-L1_prop / L1_propvar))
    )
    ratio = np.minimum(1, np.exp(postlog_prop - postlog_curr))

    u = rand()
    if u <= ratio:
        L1_curr = L1_prop
        L1_accept = True
    else:
        L1_accept = False

    return L1_curr, L1_accept


def f_updatev_pw(
    v_coef_curr,
    v_basis,
    sigma_curr,
    q1,
    q2,
    nll_cur,
    g_cur,
    SSE_curr,
    propose_v_coef,
    d_basis,
    cholC,
    h,
    L,
):
    v_coef_prop = propose_v_coef(v_coef_curr)

    q = v_coef_prop.copy()
    D = q.shape[0]
    rth = np.sqrt(h)
    g_cur = -1 * g_cur

    # sample velocity
    v = normal(size=D)
    v = cholC.T @ v

    # natural gradient
    halfng = cholC @ g_cur
    ng = cholC @ halfng

    # accumulate the power of force
    pow = rth / 2 * (np.dot(g_cur, v))

    # calculate current energy
    E_cur = nll_cur - h / 8 * (np.dot(halfng, halfng))

    randL = int(np.ceil(rand() * L))

    # Alternate full sth for position and velocity
    for l in range(0, randL):
        # Make a half step for velocity
        v = v + rth / 2 * ng

        # Make a full step for position
        rot = (q + 1j * v) * np.exp(-1j * rth)
        q = np.real(rot)
        v = np.imag(rot)

        # update geometry
        nll, g, SSEv = f_dlogl_pw(q, v_basis, d_basis, sigma_curr, q1, q2)
        halfng = cholC @ g
        ng = cholC.T @ halfng

        # Make a half step for velocity
        v = v + rth / 2 * ng

        # accumulate the power of force
        if l != randL:
            pow = pow + rth * (np.dot(g, v))

    # accumulate the power of force
    pow = pow + rth / 2 * (np.dot(g, v))

    # evaluate energy at start and end of trajectory
    E_prop = nll - h / 8 * (np.dot(halfng, halfng))

    # Accept or reject the state at end of trajectory, returning either
    # the position at the end of the trajectory or the initial position
    logRatio = -E_prop + E_cur - pow

    if (not np.isinf(logRatio)) and (np.log(rand()) < np.minimum(0, logRatio)):
        v_coef_curr = q
        theta_accept = True
        SSE_curr = SSEv
    else:
        nll = nll_cur
        g = g_cur
        theta_accept = False

    return v_coef_curr, nll, g, SSE_curr, theta_accept


def f_SSEg_pw(g, q1, q2):
    obs_domain = np.linspace(0, 1, g.shape[0])
    exp1g_temp = uf.f_predictfunction(uf.f_exp1(g), obs_domain, 0)
    pt = np.insert(bay.bcuL2norm2(obs_domain, exp1g_temp), 0, 0)
    tmp = uf.f_predictfunction(q2, pt, 0)
    vec = (q1 - tmp * exp1g_temp) ** 2
    out = vec.sum()
    return out


def f_logl_pw(g, q1, q2, var1, SSEg):
    if SSEg == 0:
        SSEg = f_SSEg_pw(g, q1, q2)

    n = q1.shape[0]
    out = (
        n * np.log(1 / np.sqrt(2 * np.pi))
        - n * np.log(np.sqrt(var1))
        - SSEg / (2 * var1)
    )

    return out


def f_updateg_pw(g_coef_curr, g_basis, var1_curr, q1, q2, SSE_curr, propose_g_coef):
    g_coef_prop = propose_g_coef(g_coef_curr)

    tst = uf.f_exp1(uf.f_basistofunction(g_basis["x"], 0, g_coef_prop["prop"], g_basis))

    while tst.min() < 0:
        g_coef_prop = propose_g_coef(g_coef_curr)
        tst = uf.f_exp1(
            uf.f_basistofunction(g_basis["x"], 0, g_coef_prop["prop"], g_basis)
        )

    if SSE_curr == 0:
        SSE_curr = f_SSEg_pw(
            uf.f_basistofunction(g_basis["x"], 0, g_coef_curr, g_basis), q1, q2
        )

    SSE_prop = f_SSEg_pw(
        uf.f_basistofunction(g_basis["x"], 0, g_coef_prop["prop"], g_basis), q1, q2
    )

    logl_curr = f_logl_pw(
        uf.f_basistofunction(g_basis["x"], 0, g_coef_curr, g_basis),
        q1,
        q2,
        var1_curr,
        SSE_curr,
    )

    logl_prop = f_logl_pw(
        uf.f_basistofunction(g_basis["x"], 0, g_coef_prop["prop"], g_basis),
        q1,
        q2,
        var1_curr,
        SSE_prop,
    )

    ratio = np.minimum(1, np.exp(logl_prop - logl_curr))

    u = np.random.rand()
    if u <= ratio:
        g_coef = g_coef_prop["prop"]
        logl = logl_prop
        SSE = SSE_prop
        accept = True
        zpcnInd = g_coef_prop["ind"]

    if u > ratio:
        g_coef = g_coef_curr
        logl = logl_curr
        SSE = SSE_curr
        accept = False
        zpcnInd = g_coef_prop["ind"]

    return g_coef, logl, SSE, accept, zpcnInd



"""
Author: Kanad Khanna
2021-11-24
Translation of MATLAB interpolation code to Python.
"""

import numpy as np
import scipy.interpolate
import scipy.integrate


def interparc(t, px, py, *args, method="spline"):
    """
    % usage: pt, f_pt = interparc(t,px,py)    % a 2-d curve
    % usage: pt, f_pt = interparc(t,px,py,pz) % a 3-d curve
    % usage: pt, f_pt = interparc(t,px,py,pz,pw,...) % a 4-d or higher dimensional curve
    % usage: pt, f_pt = interparc(t,px,py,method) % a 2-d curve, method is specified
    %
    % Interpolates new points at any fractional point along
    % the curve defined by a list of points in 2 or more
    % dimensions. The curve may be defined by any sequence
    % of non-replicated points.
    %
    % arguments: (input)
    %  t   - vector of numbers, 0 <= t <= 1, that define
    %        the fractional distance along the curve to
    %        interpolate the curve at. t = 0 will generate
    %        the very first point in the point list, and
    %        t = 1 yields the last point in that list.
    %        Similarly, t = 0.5 will yield the mid-point
    %        on the curve in terms of arc length as the
    %        curve is interpolated by a parametric spline.
    %
    %        If t is a scalar integer, at least 2, then
    %        it specifies the number of equally spaced
    %        points in arclength to be generated along
    %        the curve.
    %
    %  px, py, pz, ... - vectors of length n, defining
    %        points along the curve. n must be at least 2.
    %        Exact Replicate points should not be present
    %        in the curve, although there is no constraint
    %        that the curve has replicate independent
    %        variables.
    %
    %  method - (OPTIONAL) string flag - denotes the method
    %        used to compute the points along the curve.
    %        may be 'linear' or 'spline'
    %
    %        method == 'linear' --> Uses a linear chordal
    %               approximation to interpolate the curve.
    %               This method is the most efficient.
    %        method == 'spline' --> Uses a parametric spline
    %               approximation for the interpolation in
    %               arc length. Generally for a smooth curve,
    %               this method may be most accurate.
    %
    %        DEFAULT: 'spline'
    %
    %
    % arguments: (output)
    %  pt - Interpolated points at the specified fractional
    %        distance (in arc length) along the curve.
    %  f_pt - Evaluation of the function at the resampled points
    % Author: John D'Errico
    % e-mail: woodchips@rochester.rr.com
    % Release: 1.0
    % Release date: 3/15/2010
    """
    # Method may be any of {'linear', 'spline'}
    if method not in ("linear", "spline"):
        raise ValueError(
            "INTERPARC: Incorrect method. Method must be either 'linear' or 'spline'."
        )

    # If t is an integer, generate t evenly spaced points in [0,1]
    if isinstance(t, int) and t > 1:
        t = np.linspace(0, 1, t)
    # Otherwise, create a column vector from the array that was passed
    else:
        t = np.array(t).reshape(-1, 1)

    # Create column vectors from px and py, as well
    px = np.array(px).reshape(-1, 1)
    py = np.array(py).reshape(-1, 1)

    # Make sure that only valid t's were passed
    if (t < 0).any() or (t > 1).any():
        raise ValueError(
            "ARCLENGTH: Improper t. All elements of t must be 0 <= t <= 1."
        )

    # How many points will be interpolated?
    nt = len(t)

    # The number of points on the curve itself
    n = len(px)

    # Are px and py both vectors of the same length?
    if len(px) != len(py):
        raise ValueError(
            "ARCLENGTH: Improper px or py. px and py must be vectors of the same length."
        )
    if len(px) < 2:
        raise ValueError(
            "ARCLENGTH: Improper px or py. px and py must be vectors of length at least 2."
        )

    # Compose px and py into a single array. This way, if more dimensions are
    # provided, the extension is trivial.
    pxy = np.hstack([px, py])  # indexing is pxy[timestep, dimension]
    ndim = 2

    # Are there any other arguments?
    if len(args) > 0:
        # Create column vectors from pz, pw, etc.
        args = [np.array(p_other).reshape(-1, 1) for p_other in args]
        # Stack all dimensions into a single matrix.
        for arg in args:
            pxy = np.column_stack([pxy, arg])
        ndim = pxy.shape[1]

    # Pre-allocate the result, pt
    # pt = np.full((nt, ndim), np.NaN)

    # Compute the chordal (linear) arclength of each segment. This will be
    # needed for any of the methods.
    chordlen = np.linalg.norm(np.diff(pxy, axis=0), axis=1)

    # Normalize the arclengths to a unit total
    chordlen = chordlen / chordlen.sum()

    # Cumulative arclength
    cumarc = np.concatenate(([0], np.cumsum(chordlen)))

    # The linear method (trivial case)
    if method == "linear":
        # Which interval did each point fall in, in terms of t?
        tbins = np.digitize(t, cumarc)

        # Catch any problems at the ends
        tbins[(tbins <= 0) | (t <= 0)] = 1
        tbins[(tbins >= n) | (t >= 1)] = n - 1

        # Convert indices to 0-based indexing
        tbins = tbins - 1

        # Interpolate
        s = (t - cumarc[tbins]) / chordlen[tbins]

        # Be nice, and allow the code to work on older releases that don't have bsxfun
        # NOTE: Actually, numpy does automatic broadcasting :) Just add a dimension to s.
        pt = pxy[tbins, :] + (pxy[(tbins + 1), :] - pxy[tbins, :]) * np.expand_dims(
            s, 1
        )

        # Create the spline as a piecewise linear function
        # spl = []
        # for i in range(ndim):
        #     coefs = [np.diff(pxy[:,i]) / np.diff(cumarc),   pxy[:-1, i]]
        #     coefs = np.vstack(coefs)
        #     spl.append(scipy.interpolate.PPoly(c=coefs, x=cumarc))

        # # Evaluate the interpolated points
        # f_pt = np.vstack([f_i(pt) for f_i in spl]).T

        return np.squeeze(pt)

    # Spline method
    elif method == "spline":
        spl = []
        for i in range(ndim):
            # Compute parametric splines
            spl.append(scipy.interpolate.CubicSpline(x=cumarc, y=pxy[:, i]))
            nc = spl[i].c.size
            if nc < 4:
                # Just pretend it has cubic segments
                # NOTE: The matlab code just edits a struct. This is an object and
                # may not behave the same, so we probably need to construct a new
                # object.
                raise NotImplementedError(
                    "Fewer than 4 coefficients created. Kanad has not implemented this case yet."
                )

        # And now differentiate them
        spld = [spline.derivative() for spline in spl]

        # Catch the case where there were exactly 3 points in the curve, and
        # spline was used to generate the interpolant. In this case, spline
        # creates a curve with only 1 piece, not 2.
        if cumarc.size == 3:
            cumarc = spl[0].x
            n = cumarc.size
            chordlen = chordlen.sum()

        # Generate the total arclength along the curve by integrating each segment
        # and summing the results. The integration scheme does its job using an
        # ODE solver.
        seglen = np.zeros((n - 1, 1))
        for i in range(n - 1):
            # Extract polynomials for the derivatives.
            # NOTE: scipy.CubicSpline stores its coefs in reverse (descending) order
            polyarray = [
                np.polynomial.polynomial.Polynomial(coef=spld[j].c[:, i][::-1])
                for j in range(ndim)
            ]

            # Integrate the arclength for the i'th segment using ode45 for the integral.
            # I could have done this part with quad too, but then it would not have been
            # perfectly (numerically) consistent with the next operation in this tool.
            results = scipy.integrate.solve_ivp(
                lambda t, y: np.sqrt(
                    np.sum(np.square([P(t) for P in polyarray]), axis=0)
                ),  # sqrt((dx/dt)^2 + (dy/dt)^2 + ...)
                t_span=(0, chordlen[i]),
                y0=[
                    0,
                ],
                rtol=1.0e-9,
            )
            seglen[i] = results.y[:, -1]

        # and normalize the segments to have unit total length
        totalsplinelength = seglen.sum()
        cumseglen = np.concatenate(([0], seglen.cumsum()))

        # Which interval did each point fall into, in terms of t, but relative to the
        # cumulative arc lengths along the parametric spline?
        tbins = np.digitize(t * totalsplinelength, cumseglen)

        # Catch any problems at the ends
        tbins[(tbins <= 0) | (t <= 0)] = 1
        tbins[(tbins >= n) | (t >= 1)] = n - 1

        # Convert indices to 0-based indexing
        tbins = tbins - 1

        # Do the fractional integration within each segment for the interpolated
        # points. t is the parameter used to define the splines. It is defined in
        # terms of a linear chordal arclength. This works nicely when a linear
        # piecewise interpolant was used. However, what is asked for is an arclength
        # interpolation in terms of arclength of the spline itself. Call s the
        # arclength traveled along the spline.
        s = totalsplinelength * t

        # The ode45 solver will now include an events callback so we can catch zero crossings
        ode_events = lambda t, y: y
        ode_events.terminal = True
        ode_events.direction = 1

        ti = t.copy()
        for i in range(nt):
            # si is the piece of arc length that we will look for in this spline segment
            si = np.squeeze(s[i] - cumseglen[tbins[i]])

            # Extract polynomials for the derivatives in the interval the point lies in
            # NOTE: scipy.CubicSpline stores its coefs in reverse (descending) order
            polyarray = [
                np.polynomial.polynomial.Polynomial(
                    coef=np.squeeze(spld[j].c[:, tbins[i]][::-1])
                )
                for j in range(ndim)
            ]

            # We need to integrate in t, until the integral crosses the specified value of
            # si. Because we have defined totalsplinelength, the lengths will be normalized
            # at this point to a unit length.
            #
            # Start the ode solver at -si, so we will just look for an event where y crosses 0.
            results = scipy.integrate.solve_ivp(
                lambda t, y: np.sqrt(
                    np.sum(np.square([P(t) for P in polyarray]), axis=0)
                ),  # sqrt((dx/dt)^2 + (dy/dt)^2 + ...)
                t_span=(0, chordlen[tbins[i]]),
                y0=[
                    -si,
                ],
                rtol=1.0e-9,
                events=ode_events,
            )
            tout = results.t
            yout = results.y.flatten()  # Will this be a problem? Assuming k==1 always.
            te = results.t_events[0]

            # We only need that point where a zero crossing occurred.
            # If no crossing was found, then we can look at each end.
            if len(te) > 0:
                ti[i] = te[0] + cumarc[tbins[i]]
            else:
                # A crossing must have happened at the very beginning or the
                # end, and the ode solver missed it, not trapping that event.
                if np.abs(yout[0]) < np.abs(yout[-1]):
                    # The event must have been at the start.
                    ti[i] = tout[0] + cumarc[tbins[i]]
                else:
                    # The event must have been at the end.
                    ti[i] = tout[-1] + cumarc[tbins[i]]

        # Interpolate the parametric splines at ti to get our interpolated value.
        pt = np.hstack([f(ti) for f in spl])

        return pt


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Example 1:
    # Interpolate a set of unequally spaced points around the perimeter of
    # a unit circle, generating equally spaced points around the perimeter
    # theta = np.sort(np.random.rand(15) * 2 * np.pi)
    # theta = np.append(theta, theta[0])
    theta = np.array(
        [
            0.49196421,
            1.30404303,
            1.35277569,
            1.66329111,
            1.87935206,
            2.83274271,
            3.39728439,
            3.50759802,
            3.87032075,
            4.43891395,
            4.55715715,
            5.04428418,
            5.18313692,
            5.59810929,
            5.85648442,
            0.49196421,
        ]
    )
    px = np.cos(theta)
    py = np.sin(theta)

    # Interpolate using parametric splines
    pt, _ = interparc(100, px, py, "spline")

    # Plot the result
    fig, ax = plt.subplots()
    ax.plot(px, py, "r*")
    ax.plot(pt[:, 0], pt[:, 1], "b-o")
    ax.set_ylim(-1.1, 1.1)
    ax.set_xlim(-1.1, 1.1)
    ax.grid()
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_title("Points in blue are uniform in arclength around the circle")
    plt.show()



"""
Functional Tolerance Bounds using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.fPCA as fpca
from scipy.stats import chi2
from numpy.linalg import eig
from fdasrsf.boxplots import ampbox, phbox


def bootTB(f, time, a=0.05, p=0.99, B=500, no=5, parallel=True):
    """
    This function computes tolerance bounds for functional data containing
    phase and amplitude variation using bootstrap sampling

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size M describing the sample points
    :param a: confidence level of tolerance bound (default = 0.05)
    :param p: coverage level of tolerance bound (default = 0.99)
    :param B: number of bootstrap samples (default = 500)
    :param no: number of principal components (default = 5)
    :param parallel: enable parallel processing (default = T)
    :type f: np.ndarray
    :type time: np.ndarray

    :rtype: tuple of boxplot objects
    :return amp: amplitude tolerance bounds
    :rtype out_med: ampbox object
    :return ph: phase tolerance bounds
    :rtype out_med: phbox object
    :return out_med: alignment results
    :rtype out_med: fdawarp object

    """
    eps = np.finfo(np.double).eps
    (M, N) = f.shape

    # Align Data
    out_med = fs.fdawarp(f, time)
    out_med.srsf_align(method="median", parallel=parallel)

    # Calculate CI
    # a% tolerance bound with p%
    fn = out_med.fn
    qn = out_med.qn
    gam = out_med.gam
    q0 = out_med.q0
    print("Bootstrap Sampling")
    bootlwr_amp = np.zeros((M, B))
    bootupr_amp = np.zeros((M, B))
    bootlwr_ph = np.zeros((M, B))
    bootupr_ph = np.zeros((M, B))
    for k in range(B):
        out_med.joint_gauss_model(n=100, no=no)
        obja = ampbox(out_med)
        obja.construct_boxplot(1 - p, 0.3)
        objp = phbox(out_med)
        objp.construct_boxplot(1 - p, 0.3)
        bootlwr_amp[:, k] = obja.Q1a
        bootupr_amp[:, k] = obja.Q3a
        bootlwr_ph[:, k] = objp.Q1a
        bootupr_ph[:, k] = objp.Q3a

    # tolerance bounds
    boot_amp = np.hstack((bootlwr_amp, bootupr_amp))
    f, g, g2 = uf.gradient_spline(time, boot_amp, False)
    boot_amp_q = g / np.sqrt(abs(g) + eps)
    boot_ph = np.hstack((bootlwr_ph, bootupr_ph))
    boot_out = out_med
    boot_out.fn = boot_amp
    boot_out.qn = boot_amp_q
    boot_out.gam = boot_ph
    boot_out.rsamps = False
    amp = ampbox(boot_out)
    amp.construct_boxplot(a, 0.3)
    ph = phbox(boot_out)
    ph.construct_boxplot(a, 0.3)

    return amp, ph, out_med


def pcaTB(f, time, a=0.5, p=0.99, no=5, parallel=True):
    """
    This function computes tolerance bounds for functional data containing
    phase and amplitude variation using fPCA

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size M describing the sample points
    :param a: confidence level of tolerance bound (default = 0.05)
    :param p: coverage level of tolerance bound (default = 0.99)
    :param no: number of principal components (default = 5)
    :param parallel: enable parallel processing (default = T)
    :type f: np.ndarray
    :type time: np.ndarray

    :rtype: tuple of boxplot objects
    :return warp: alignment data from time_warping
    :return pca: functional pca from jointFPCA
    :return tol: tolerance factor

    """

    # Align Data
    out_warp = fs.fdawarp(f, time)
    out_warp.srsf_align(method="median", parallel=parallel)

    # Calculate pca
    out_pca = fpca.fdajpca(out_warp)
    out_pca.calc_fpca(no)

    # Calculate TB
    tol = mvtol_region(out_pca.coef, a, p, 100000)

    return out_warp, out_pca, tol


def mvtol_region(x, alpha, P, B):
    """
    Computes tolerance factor for multivariate normal

    Krishnamoorthy, K. and Mondal, S. (2006), Improved Tolerance Factors for Multivariate Normal
    Distributions, Communications in Statistics - Simulation and Computation, 35, 461–478.

    :param x: (M,N) matrix defining N variables of M samples
    :param alpha: confidence level
    :param P: coverage level
    :param B: number of bootstrap samples

    :rtype: double
    :return tol: tolerance factor

    """
    n, p = x.shape

    q_squared = chi2.rvs(1, size=(p, B)) / n
    L = np.zeros((p, B))
    for k in range(B):
        L[:, k] = eig(rwishart(n - 1, p))[0]

    c1 = (1 + q_squared) / L
    c1 = c1.sum()
    c2 = (1 + 2 * q_squared) / (L**2)
    c2 = c2.sum()
    c3 = (1 + 3 * q_squared) / (L**3)
    c3 = c3.sum()
    a = (c2**3) / (c3**2)
    T = (n - 1) * (np.sqrt(c2 / a) * (chi2.ppf(P, a) - a) + c1)
    tol = np.quantile(T, 1 - alpha)

    return tol


def rwishart(df, p):
    """
    Computes a random wishart matrix

    :param df: degree of freedom
    :param p: number of dimensions

    :rtype: double
    :return R: matrix

    """
    R = np.zeros((p, p))
    R = R.flatten()
    R[:: p + 1] = np.sqrt(chi2.rvs(np.arange(df, df - p, -1), size=p))
    if p > 1:
        pseq = np.arange(0, p)
        tmp = [np.arange(0, x + 1) for x in np.arange(0, p - 1)]
        R[np.repeat(p * pseq, pseq) + np.concatenate(tmp).ravel()] = np.random.randn(
            int(p * (p - 1) / 2)
        )

    R = R.reshape((p, p))
    R = np.matmul(R.T, R)

    return R



"""
A python package for functional data analysis using the square root
slope framework which performs pair-wise and group-wise
alignment as well as modeling using functional component
analysis

"""
__all__ = [
    "time_warping",
    "utility_functions",
    "bayesian_functions",
    "curve_stats",
    "geodesic",
    "curve_functions",
    "geometry",
    "pcr_regression",
    "tolerance",
    "boxplots",
    "curve_regression",
    "regression",
    "fPCA",
    "elastic_changepoint",
    "elastic_glm_regression",
    "curve_pcr_regression",
    "kmeans",
    "image",
    "image_functions",
    "gp",
]

__version__ = "2.4.2"

import sys

if sys.version_info[0] == 3 and sys.version_info[1] < 6:
    raise ImportError("Python Version 3.6 or above is required for fdasrsf.")
else:  # Python 3
    pass
    # Here we can also check for specific Python 3 versions, if needed

del sys

from .time_warping import (
    fdawarp,
    align_fPCA,
    align_fPLS,
    pairwise_align_bayes,
    pairwise_align_functions,
)
from .time_warping import pairwise_align_bayes_infHMC
from .plot_style import (
    f_plot,
    rstyle,
    plot_curve,
    plot_reg_open_curve,
    plot_geod_open_curve,
    plot_geod_close_curve,
)
from .utility_functions import (
    smooth_data,
    optimum_reparam,
    f_to_srsf,
    gradient_spline,
    elastic_distance,
    invertGamma,
    srsf_to_f,
)
from .utility_functions import (
    SqrtMean,
    SqrtMeanInverse,
    cumtrapzmid,
    rgam,
    outlier_detection,
    innerprod_q,
    SqrtMedian,
)
from .utility_functions import (
    optimum_reparam_pair,
    warp_q_gamma,
    resamplefunction,
    warp_f_gamma,
    elastic_depth,
)
from .fPCA import fdavpca, fdahpca, fdajpca, fdajpcah
from .fPLS import pls_svd
from .regression import elastic_logistic, elastic_regression, elastic_mlogistic
from .elastic_glm_regression import elastic_glm_regression
from .pcr_regression import (
    elastic_pcr_regression,
    elastic_lpcr_regression,
    elastic_mlpcr_regression,
)
from .curve_pcr_regression import elastic_curve_pcr_regression
from .elastic_changepoint import (
    elastic_amp_change_ff,
    elastic_change,
    elastic_ph_change_ff,
)
from .boxplots import ampbox, phbox
from .tolerance import bootTB, pcaTB
from .kmeans import kmeans_align
from .image import reparam_image
from .image_functions import formbasisTid, apply_gam_imag, makediffeoid
from .curve_functions import (
    resamplecurve,
    calculatecentroid,
    curve_to_q,
    optimum_reparam_curve,
    find_best_rotation,
    elastic_distance_curve,
)
from .curve_functions import q_to_curve, find_rotation_and_seed_coord
from .curve_stats import fdacurve
from .curve_regression import (
    oc_elastic_regression,
    oc_elastic_logistic,
    preproc_open_curve,
    oc_elastic_mlogistic,
)
from .geometry import inv_exp_map, exp_map
from .geodesic import geod_sphere, path_straightening
from .umap_metric import efda_distance, efda_distance_curve
from .rbfgs import rlbfgs



import matplotlib
import pylab
import fdasrsf.curve_functions as cf
from numpy import tile, array, arange
import matplotlib.pyplot as plt


def rstyle(ax, pres=False):
    """
    Styles x,y axes to appear like ggplot2
    Must be called after all plot and axis manipulation operations have been
    carried out (needs to know final tick spacing)
    """
    # Set the style of the major and minor grid lines, filled blocks
    if pres:
        ax.grid(True, "major", color="w", linestyle="-", linewidth=0.7)
        ax.grid(True, "minor", color="0.99", linestyle="-", linewidth=0.4)
    else:
        ax.grid(True, "major", color="w", linestyle="-", linewidth=1.4)
        ax.grid(True, "minor", color="0.99", linestyle="-", linewidth=0.7)
    ax.patch.set_facecolor("0.90")
    ax.set_axisbelow(True)

    # Set minor tick spacing to 1/2 of the major ticks
    ax.xaxis.set_minor_locator(
        (pylab.MultipleLocator((plt.xticks()[0][1] - plt.xticks()[0][0]) / 2.0))
    )
    ax.yaxis.set_minor_locator(
        (pylab.MultipleLocator((plt.yticks()[0][1] - plt.yticks()[0][0]) / 2.0))
    )

    # Remove axis border
    for child in ax.get_children():
        if isinstance(child, matplotlib.spines.Spine):
            child.set_alpha(0)

    # Restyle the tick lines
    for line in ax.get_xticklines() + ax.get_yticklines():
        if pres:
            line.set_markersize(4)
            line.set_color("gray")
            line.set_markeredgewidth(0.8)
        else:
            line.set_markersize(5)
            line.set_color("gray")
            line.set_markeredgewidth(1.4)

    # Remove the minor tick lines
    for line in ax.xaxis.get_ticklines(minor=True) + ax.yaxis.get_ticklines(minor=True):
        line.set_markersize(0)

    # Only show bottom left ticks, pointing out of axis
    plt.rcParams["xtick.direction"] = "out"
    plt.rcParams["ytick.direction"] = "out"
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")


def rstyle_bw(ax):
    """
    Styles x,y axes to appear like ggplot2
    Must be called after all plot and axis manipulation operations have been
    carried out (needs to know final tick spacing)
    """
    # Set the style of the major and minor grid lines, filled blocks
    ax.grid(True, "major", color="0.88", linestyle="-", linewidth=0.7)
    ax.grid(True, "minor", color="0.95", linestyle="-", linewidth=0.4)
    ax.set_axisbelow(True)

    # Set minor tick spacing to 1/2 of the major ticks
    ax.xaxis.set_minor_locator(
        (pylab.MultipleLocator((plt.xticks()[0][1] - plt.xticks()[0][0]) / 2.0))
    )
    ax.yaxis.set_minor_locator(
        (pylab.MultipleLocator((plt.yticks()[0][1] - plt.yticks()[0][0]) / 2.0))
    )

    # Remove axis border
    ax.spines["top"].set_alpha(0)
    ax.spines["right"].set_alpha(0)

    # Restyle the tick lines
    for line in ax.get_xticklines() + ax.get_yticklines():
        line.set_markersize(4)
        #     line.set_color("gray")
        line.set_markeredgewidth(0.8)

    # Remove the minor tick lines
    for line in ax.xaxis.get_ticklines(minor=True) + ax.yaxis.get_ticklines(minor=True):
        line.set_markersize(0)

    # Only show bottom left ticks, pointing out of axis
    plt.rcParams["xtick.direction"] = "out"
    plt.rcParams["ytick.direction"] = "out"
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")


def f_plot(time, f, title="Data", bw=False, pres=False):
    """
    plots function data using matplotlib

    :param time: vector of size N describing the sample points
    :param f: numpy ndarray of shape (M,N) of M SRSFs with N samples
    :param title: string of title

    :return fig: figure definition
    :return ax: axes definition

    """

    fig, ax = plt.subplots()
    ax.plot(time, f)
    plt.title(title)
    plt.style.use("ggplot")

    return fig, ax


def plot_curve(beta):
    """
    plots curve

    :param beta: numpy array of shape (2,M) of M samples

    :return fig: figure defintion
    :return ax: axes
    """
    fig, ax = plt.subplots()
    ax.plot(beta[0, :], beta[1, :], "r", linewidth=2)
    ax.set_aspect("equal")
    ax.axis("off")

    return fig, ax


def plot_reg_open_curve(beta1, beta2n):
    """
    plots registration between two open curves using matplotlib

    :param beta: numpy ndarray of shape (2,M) of M samples
    :param beta: numpy ndarray of shape (2,M) of M samples

    :return fig: figure definition
    :return ax: axes definition

    """
    T = beta1.shape[1]
    centroid1 = cf.calculatecentroid(beta1)
    beta1 = beta1 - tile(centroid1, [T, 1]).T
    centroid2 = cf.calculatecentroid(beta2n)
    beta2n = beta2n - tile(centroid2, [T, 1]).T
    beta2n[0, :] = beta2n[0, :] + 1.3
    beta2n[1, :] = beta2n[1, :] - 0.1

    fig, ax = plt.subplots()
    ax.plot(beta1[0, :], beta1[1, :], "r", linewidth=2)
    fig.hold()
    ax.plot(beta2n[0, :], beta2n[1, :], "b-o", linewidth=2)

    for j in range(0, int(T / 5)):
        i = j * 5
        ax.plot(
            array([beta1[0, i], beta2n[0, i]]),
            array([beta1[1, i], beta2n[1, i]]),
            "k",
            linewidth=1,
        )

    ax.set_aspect("equal")
    ax.axis("off")
    fig.hold()

    return fig, ax


def plot_geod_open_curve(PsiX):
    """
    plots geodesic between two open curves using matplotlib

    :param PsiX: numpy ndarray of shape (2,M,k) of M samples

    :return fig: figure definition
    :return ax: axes definition

    """
    k = PsiX.shape[2]
    fig, ax = plt.subplots()
    fig.hold()
    for tau in range(0, k):
        ax.plot(0.35 * tau + PsiX[0, :, tau], PsiX[1, :, tau], "k", linewidth=2)

    ax.set_aspect("equal")
    ax.axis("off")
    fig.hold()

    return fig, ax


def plot_geod_close_curve(pathsqnc):
    """
    plots geodesic between two closed curves using matplotlib

    :param pathsqnc: numpy ndarray of shape (2,M,k,i) of M samples

    :return fig: figure definition
    :return ax: axes definition

    """
    i = pathsqnc.shape[3]
    k = pathsqnc.shape[2]
    if i > 1:
        plotidx = arange(0, i)
        fig, ax = plt.subplots(plotidx.size, k, sharex=True, sharey=True)
        for j in plotidx:
            for tau in range(0, k):
                beta_tmp = pathsqnc[:, :, tau, j]
                ax[j, tau].plot(beta_tmp[0, :], beta_tmp[1, :], "r", linewidth=2)
                ax[j, tau].set_aspect("equal")
                ax[j, tau].axis("off")
    else:
        fig, ax = plt.subplots(1, k, sharex=True, sharey=True)
        for tau in range(0, k):
            beta_tmp = pathsqnc[:, :, tau, j]
            ax[tau].plot(beta_tmp[0, :], beta_tmp[1, :], "r", linewidth=2)
            ax[tau].set_aspect("equal")
            ax[tau].axis("off")

    return fig, ax



import setuptools
import numpy
import sys, os
import platform
from distutils.core import setup
from distutils.core import Command
from findblas.distutils import build_ext_with_blas
from distutils.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize
from setuptools import dist
from distutils.sysconfig import get_config_var, get_python_inc
from distutils.version import LooseVersion

sys.path.insert(1, "src/")
import dp_build

# Make sure I have the right Python version.
if sys.version_info[:2] < (3, 9):
    print(
        (
            "fdasrsf requires Python 3.9 or newer. Python %d.%d detected"
            % sys.version_info[:2]
        )
    )
    sys.exit(-1)


class build_docs(Command):
    """Builds the documentation"""

    description = "builds the documentation"
    user_options = []

    def initialize_options(self):
        self.all = None

    def finalize_options(self):
        pass

    def run(self):
        import os

        os.system("sphinx-build -b html doc/source doc/build/html")
        os.system("sphinx-build -b latex doc/source doc/build/latex")
        os.system("sphinx-build -b man doc/source doc/build/man")
        os.chdir("doc/build/latex")
        os.system("latexmk -pdf fdasrsf.tex")
        os.chdir("../../../")


if sys.platform == "darwin":
    mac_ver = str(LooseVersion(get_config_var("MACOSX_DEPLOYMENT_TARGET")))
    os.environ["MACOSX_DEPLOYMENT_TARGET"] = mac_ver

extensions = [
    Extension(
        name="optimum_reparamN2",
        sources=[
            "src/optimum_reparamN2.pyx",
            "src/DynamicProgrammingQ2.c",
            "src/dp_grid.c",
            "src/dp_nbhd.c",
        ],
        include_dirs=[numpy.get_include()],
        language="c",
    ),
    Extension(
        name="fpls_warp",
        sources=["src/fpls_warp.pyx", "src/fpls_warp_grad.c", "src/misc_funcs.c"],
        include_dirs=[numpy.get_include()],
        language="c",
    ),
    Extension(
        name="mlogit_warp",
        sources=["src/mlogit_warp.pyx", "src/mlogit_warp_grad.c", "src/misc_funcs.c"],
        include_dirs=[numpy.get_include()],
        language="c",
    ),
    Extension(
        name="ocmlogit_warp",
        sources=[
            "src/ocmlogit_warp.pyx",
            "src/ocmlogit_warp_grad.c",
            "src/misc_funcs.c",
        ],
        include_dirs=[numpy.get_include()],
        language="c",
    ),
    Extension(
        name="oclogit_warp",
        sources=["src/oclogit_warp.pyx", "src/oclogit_warp_grad.c", "src/misc_funcs.c"],
        include_dirs=[numpy.get_include()],
        language="c",
    ),
    Extension(
        name="optimum_reparam_N",
        sources=["src/optimum_reparam_N.pyx", "src/DP.c"],
        include_dirs=[numpy.get_include()],
        language="c",
    ),
    Extension(
        name="cbayesian",
        sources=["src/cbayesian.pyx", "src/bayesian.cpp"],
        include_dirs=[numpy.get_include()],
        language="c++",
        extra_compile_args=["-std=c++11"],
    ),
    Extension(
        name="crbfgs",
        sources=["src/crbfgs.pyx", "src/rbfgs.cpp"],
        include_dirs=[numpy.get_include()],
        language="c++",
        extra_compile_args=["-std=c++11"],
    ),
    Extension(
        name="cimage",
        sources=["src/imagecpp.pyx", "src/UnitSquareImage.cpp"],
        include_dirs=[numpy.get_include()],
        language="c++",
    ),
    dp_build.ffibuilder.distutils_extension(),
]


setup(
    cmdclass={"build_ext": build_ext_with_blas, "build_docs": build_docs},
    ext_modules=extensions,
    name="fdasrsf",
    version="2.5.14",
    packages=["fdasrsf"],
    url="http://research.tetonedge.net",
    license="LICENSE.txt",
    author="J. Derek Tucker",
    author_email="jdtuck@sandia.gov",
    scripts=["bin/ex_srsf_align.py"],
    keywords=["functional data analysis"],
    description="functional data analysis using the square root slope framework",
    long_description=open("README.md", encoding="utf8").read(),
    data_files=[("share/man/man1", ["doc/build/man/fdasrsf.1"])],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
)



"""
Warping Invariant PCR Regression using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.fPCA as fpca
import fdasrsf.regression as rg
import fdasrsf.geometry as geo
from scipy.linalg import inv
from scipy.stats import t
from scipy.optimize import fmin_l_bfgs_b


class elastic_pcr_regression:
    """
    This class provides elastic pcr regression for functional data using the
    SRVF framework accounting for warping

    Usage:  obj = elastic_pcr_regression(f,y,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N
    :param warp_data: fdawarp object of alignment
    :param pca: class dependent on fPCA method used object of fPCA
    :param alpha: intercept
    :param b: coefficient vector
    :param SSE: sum of squared errors

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_pcr_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception("Columns of f and time must be equal")

        self.f = f
        self.y = y
        self.time = time

    def calc_model(
        self,
        pca_method="combined",
        no=5,
        var_exp=None,
        smooth_data=False,
        sparam=25,
        parallel=False,
        C=None,
    ):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param pca_method: string specifying pca method (options = "combined",
                        "vert", or "horiz", default = "combined")
        :param no: scalar specify number of principal components (default=5)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: run in parallel (default = F)
        :param C: scale balance parameter for combined method (default = None)
        """

        if smooth_data:
            self.f = fs.smooth_data(self.f, sparam)

        N1 = self.f.shape[1]

        # Align Data
        self.warp_data = fs.fdawarp(self.f, self.time)
        self.warp_data.srsf_align(parallel=parallel)

        # Calculate PCA
        if pca_method == "combined":
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method == "vert":
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method == "horiz":
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception("Invalid fPCA Method")
        self.pca.calc_fpca(no=no, var_exp=var_exp)

        no = self.pca.no

        # OLS using PCA basis
        lam = 0
        R = 0
        Phi = np.ones((N1, no + 1))
        Phi[:, 1:(no + 1)] = self.pca.coef
        xx = np.dot(Phi.T, Phi)
        inv_xx = inv(xx + lam * R)
        xy = np.dot(Phi.T, self.y)
        b = np.dot(inv_xx, xy)
        alpha = b[0]
        b = b[1:no + 1]

        # compute the SSE
        int_X = np.zeros(N1)
        for ii in range(0, N1):
            int_X[ii] = np.sum(self.pca.coef[ii, :] * b)

        y_pred = alpha + int_X
        SSE = np.sum((self.y - y_pred) ** 2)

        self.Ntr = len(self.y)
        stdev = np.sqrt(sum((y_pred - self.y) ** 2) / (self.Ntr - 2))

        self.alpha = alpha
        self.b = b
        self.SSE = SSE
        self.stdev = stdev
        self.pca_method = pca_method

        return

    def predict(self, newdata=None, alpha=0.05):
        """
        This function performs prediction on regression model on new
        data if available or current stored data in object
        Usage:  obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs
                        the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        if newdata is not None:
            f = newdata["f"]
            n = f.shape[1]
            y = newdata["y"]
            if newdata["smooth"]:
                sparam = newdata["sparam"]
                f = fs.smooth_data(f, sparam)
            
            self.pca.project(f)
            a = self.pca.new_coef

            self.y_int = np.zeros((n, 2))
            X = inv(self.pca.coef.T @ self.pca.coef)
            df = self.Ntr - self.pca.coef.shape[1] - 1
            self.y_pred = np.zeros(n)
            for ii in range(0, n):
                self.y_pred[ii] = self.alpha + np.dot(a[ii, :], self.b)
                interval = (
                    t.ppf(alpha / 2, df)
                    * self.stdev
                    * np.sqrt(1 + a[ii, :] @ X @ a[ii, :])
                )
                self.y_int[ii, 0] = interval
                self.y_int[ii, 1] = -interval

            if y is None:
                self.SSE = np.nan
            else:
                self.SSE = np.sum((y - self.y_pred) ** 2)
        else:
            n = self.pca.coef.shape[0]
            self.y_pred = np.zeros(n)
            self.y_int = np.zeros((n, 2))
            X = inv(self.pca.coef.T @ self.pca.coef)
            df = self.Ntr - self.pca.coef.shape[1] - 1
            for ii in range(0, n):
                self.y_pred[ii] = self.alpha + np.dot(self.pca.coef[ii, :], self.b)
                interval = (
                    t.ppf(alpha / 2, df)
                    * self.stdev
                    * np.sqrt(1 + self.pca.coef[ii, :] @ X @ self.pca.coef[ii, :])
                )
                self.y_int[ii, 0] = interval
                self.y_int[ii, 1] = -interval

            self.SSE = np.sum((self.y - self.y_pred) ** 2)

        return


class elastic_lpcr_regression:
    """
    This class provides elastic logistic pcr regression for functional
    data using the SRVF framework accounting for warping

    Usage:  obj = elastic_lpcr_regression(f,y,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N (-1/1)
    :param warp_data: fdawarp object of alignment
    :param pca: class dependent on fPCA method used object of fPCA
    :param information:
    :param alpha: intercept
    :param b: coefficient vector
    :param Loss: logistic loss
    :param PC: probability of classification
    :param ylabels: predicted labels

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_lpcr_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception("Columns of f and time must be equal")

        self.f = f
        self.y = y
        self.time = time

    def calc_model(
        self, pca_method="combined", no=5, smooth_data=False, sparam=25, parallel=False
    ):
        """
        This function identifies a logistic regression model with phase-variability
        using elastic pca

        :param pca_method: string specifying pca method (options = "combined",
                        "vert", or "horiz", default = "combined")
        :param no: scalar specify number of principal components (default=5)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: calculate in parallel (default = F)
        :type f: np.ndarray
        :type time: np.ndarray
        """

        if smooth_data:
            self.f = fs.smooth_data(self.f, sparam)

        N1 = self.f.shape[1]

        # Align Data
        self.warp_data = fs.fdawarp(self.f, self.time)
        self.warp_data.srsf_align(parallel=parallel)

        # Calculate PCA
        if pca_method == "combined":
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method == "vert":
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method == "horiz":
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception("Invalid fPCA Method")
        self.pca.calc_fpca(no)

        # OLS using PCA basis
        lam = 0
        R = 0
        Phi = np.ones((N1, no + 1))
        Phi[:, 1 : (no + 1)] = self.pca.coef
        # Find alpha and beta using l_bfgs
        b0 = np.zeros(no + 1)
        out = fmin_l_bfgs_b(
            rg.logit_loss,
            b0,
            fprime=rg.logit_gradient,
            args=(Phi, self.y),
            pgtol=1e-10,
            maxiter=200,
            maxfun=250,
            factr=1e-30,
        )

        b = out[0]
        alpha = b[0]

        # compute the Loss
        LL = rg.logit_loss(b, Phi, self.y)

        b = b[1: no + 1]

        self.alpha = alpha
        self.b = b
        self.LL = LL
        self.pca_method = pca_method

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        if newdata is not None:
            f = newdata["f"]
            n = f.shape[1]
            y = newdata["y"]
            if newdata["smooth"]:
                sparam = newdata["sparam"]
                f = fs.smooth_data(f, sparam)

            self.pca.project(f)
            a = self.pca.new_coef

            self.y_pred = np.zeros(n)
            for ii in range(0, n):
                self.y_pred[ii] = self.alpha + np.sum(a[ii, :] * self.b)

            if y is None:
                self.y_pred = rg.phi(self.y_pred)
                self.y_labels = np.ones(n)
                self.y_labels[self.y_pred < 0.5] = -1
                self.PC = np.nan
            else:
                self.y_pred = rg.phi(self.y_pred)
                self.y_labels = np.ones(n)
                self.y_labels[self.y_pred < 0.5] = -1
                TP = np.sum(y[self.y_labels == 1] == 1)
                FP = np.sum(y[self.y_labels == -1] == 1)
                TN = np.sum(y[self.y_labels == -1] == -1)
                FN = np.sum(y[self.y_labels == 1] == -1)
                self.PC = (TP + TN) / (TP + FP + FN + TN)
        else:
            n = self.pca.coef.shape[0]
            self.y_pred = np.zeros(n)
            for ii in range(0, n):
                self.y_pred[ii] = self.alpha + np.dot(self.pca.coef[ii, :], self.b)

            self.y_pred = rg.phi(self.y_pred)
            self.y_labels = np.ones(n)
            self.y_labels[self.y_pred < 0.5] = -1
            TP = np.sum(self.y[self.y_labels == 1] == 1)
            FP = np.sum(self.y[self.y_labels == -1] == 1)
            TN = np.sum(self.y[self.y_labels == -1] == -1)
            FN = np.sum(self.y[self.y_labels == 1] == -1)
            self.PC = (TP + TN) / (TP + FP + FN + TN)

        return


class elastic_mlpcr_regression:
    """
    This class provides elastic multinomial logistic pcr regression for functional
    data using the SRVF framework accounting for warping

    Usage:  obj = elastic_mlpcr_regression(f,y,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N
    :param Y: coded label matrix
    :param warp_data: fdawarp object of alignment
    :param pca: class dependent on fPCA method used object of fPCA
    :param information:
    :param alpha: intercept
    :param b: coefficient vector
    :param Loss: logistic loss
    :param PC: probability of classification
    :param ylabels: predicted labels

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_mlpcr_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception("Columns of f and time must be equal")

        self.f = f
        self.y = y
        self.time = time
        N1 = f.shape[1]

        # Code labels
        m = y.max()
        self.n_classes = m
        self.Y = np.zeros((N1, m), dtype=int)
        for ii in range(0, N1):
            self.Y[ii, y[ii] - 1] = 1

    def calc_model(
        self, pca_method="combined", no=5, smooth_data=False, sparam=25, parallel=False
    ):
        """
        This function identifies a logistic regression model with phase-variability
        using elastic pca

        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: numpy array of N responses
        :param time: vector of size M describing the sample points
        :param pca_method: string specifying pca method (options = "combined",
                        "vert", or "horiz", default = "combined")
        :param no: scalar specify number of principal components (default=5)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: run model in parallel (default = F)
        :type f: np.ndarray
        :type time: np.ndarray
        """

        if smooth_data:
            self.f = fs.smooth_data(self.f, sparam)

        N1 = self.f.shape[1]

        # Align Data
        self.warp_data = fs.fdawarp(self.f, self.time)
        self.warp_data.srsf_align(parallel=parallel)

        # Calculate PCA
        if pca_method == "combined":
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method == "vert":
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method == "horiz":
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception("Invalid fPCA Method")
        self.pca.calc_fpca(no)

        # OLS using PCA basis
        lam = 0
        R = 0
        Phi = np.ones((N1, no + 1))
        Phi[:, 1: (no + 1)] = self.pca.coef
        # Find alpha and beta using l_bfgs
        b0 = np.zeros(self.n_classes * (no + 1))
        out = fmin_l_bfgs_b(
            rg.mlogit_loss,
            b0,
            fprime=rg.mlogit_gradient,
            args=(Phi, self.Y),
            pgtol=1e-10,
            maxiter=200,
            maxfun=250,
            factr=1e-30,
        )

        b = out[0]
        B0 = b.reshape(no + 1, self.n_classes)
        alpha = B0[0, :]

        # compute the Loss
        LL = rg.mlogit_loss(b, Phi, self.y)

        b = B0[1: no + 1, :]

        self.alpha = alpha
        self.b = b
        self.LL = LL
        self.pca_method = pca_method

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        m = self.n_classes

        if newdata is not None:
            f = newdata["f"]
            n = f.shape[1]
            y = newdata["y"]
            if newdata["smooth"]:
                sparam = newdata["sparam"]
                f = fs.smooth_data(f, sparam)

            self.pca.project(f)
            a = self.pca.new_coef

            self.y_pred = np.zeros(n)
            for ii in range(0, n):
                for jj in range(0, m):
                    self.y_pred[ii, jj] = self.alpha[jj] + np.sum(
                        a[ii, :] * self.b[:, jj]
                    )

            if y is None:
                self.y_pred = rg.phi(self.y_pred.reshape((1, n * m)))
                self.y_pred = self.y_pred.reshape((n, m))
                self.y_labels = np.argmax(self.y_pred, axis=1)
                self.PC = np.nan
            else:
                self.y_pred = rg.phi(self.y_pred.reshape((1, n * m)))
                self.y_pred = self.y_pred.reshape((n, m))
                self.y_labels = np.argmax(self.y_pred, axis=1)
                self.PC = np.zeros(m)
                cls_set = np.arange(0, m)
                for ii in range(0, m):
                    cls_sub = np.setdiff1d(cls_set, ii)
                    TP = np.sum(y[self.y_labels == ii] == ii)
                    FP = np.sum(y[np.in1d(self.y_labels, cls_sub)] == ii)
                    TN = np.sum(
                        y[np.in1d(self.y_labels, cls_sub)]
                        == self.y_labels[np.in1d(self.y_labels, cls_sub)]
                    )
                    FN = np.sum(np.in1d(y[self.y_labels == ii], cls_sub))
                    self.PC[ii] = (TP + TN) / (TP + FP + FN + TN)

                self.PCo = np.sum(y == self.y_labels) / self.y_labels.shape[0]
        else:
            n = self.pca.coef.shape[0]
            self.y_pred = np.zeros((n, m))
            for ii in range(0, n):
                for jj in range(0, m):
                    self.y_pred[ii, jj] = self.alpha[jj] + np.sum(
                        self.pca.coef[ii, :] * self.b[:, jj]
                    )

            self.y_pred = rg.phi(self.y_pred.reshape((1, n * m)))
            self.y_pred = self.y_pred.reshape((n, m))
            self.y_labels = np.argmax(self.y_pred, axis=1)
            self.PC = np.zeros(m)
            cls_set = np.arange(0, m)
            for ii in range(0, m):
                cls_sub = np.setdiff1d(cls_set, ii)
                TP = np.sum(self.y[self.y_labels == ii] == ii)
                FP = np.sum(self.y[np.in1d(self.y_labels, cls_sub)] == ii)
                TN = np.sum(
                    self.y[np.in1d(self.y_labels, cls_sub)]
                    == self.y_labels[np.in1d(self.y_labels, cls_sub)]
                )
                FN = np.sum(np.in1d(y[self.y_labels == ii], cls_sub))
                self.PC[ii] = (TP + TN) / (TP + FP + FN + TN)

            self.PCo = np.sum(y == self.y_labels) / self.y_labels.shape[0]

        return



"""
Elastic Functional Boxplots

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
import numpy as np
from scipy.integrate import trapezoid
import fdasrsf.utility_functions as uf
import fdasrsf.geometry as geo
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import collections


class ampbox:
    """
    This class provides amplitude boxplot for functional data using the
    SRVF framework

    Usage:  obj = ampbox(warp_data)

    :param warp_data: fdawarp class with alignment data
    :type warp_data: fdawarp
    :param Q1: First quartile
    :param Q3: Second quartile
    :param Q1a: First quantile based on alpha
    :param Q3a: Second quantile based on alpha
    :param minn: minimum extreme function
    :param maxx: maximum extreme function
    :param outlier_index: indexes of outlier functions
    :param f_median: median function
    :param q_median: median srvf
    :param plt: surface plot mesh

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the ampbox class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        if fdawarp.type == "mean":
            raise Exception(
                'Please align fdawarp class using srsf_align with method="median"!'
            )

        self.warp_data = fdawarp

    def construct_boxplot(self, alpha=0.05, k_a=1):
        """
        This function constructs the amplitude boxplot using the elastic
        square-root slope (srsf) framework.

        :param alpha: quantile value (e.g.,=.05, i.e., 95%)
        :param k_a: scalar for outlier cutoff (e.g.,=1)
        """

        if self.warp_data.rsamps:
            ft = self.warp_data.fs
            f_median = self.warp_data.fmean
            qt = self.warp_data.qs
            q_median = self.warp_data.mqn
            time = self.warp_data.time
        else:
            ft = self.warp_data.fn
            f_median = self.warp_data.fmean
            qt = self.warp_data.qn
            q_median = self.warp_data.mqn
            time = self.warp_data.time

        N = ft.shape[1]
        lam = 0.5

        # compute amplitude distances
        dy = np.zeros(N)
        for i in range(0, N):
            dy[i] = np.sqrt(trapezoid((q_median - qt[:, i]) ** 2, time))

        dy_ordering = dy.argsort()
        CR_50 = dy_ordering[0 : np.ceil(N / 2).astype("int")]
        tmp = dy[CR_50]
        m = tmp.max()

        # identify amplitude quartiles
        angle = np.zeros((CR_50.shape[0], CR_50.shape[0]))
        energy = np.zeros((CR_50.shape[0], CR_50.shape[0]))
        for i in range(0, CR_50.shape[0] - 1):
            for j in range(i + 1, CR_50.shape[0]):
                q1 = qt[:, CR_50[i]] - q_median
                q3 = qt[:, CR_50[j]] - q_median
                q1 = q1 / np.sqrt(trapezoid(q1**2, time))
                q3 = q3 / np.sqrt(trapezoid(q3**2, time))
                angle[i, j] = trapezoid(q1 * q3, time)
                energy[i, j] = (1 - lam) * (
                    dy[CR_50[i]] / m + dy[CR_50[j]] / m
                ) - lam * (angle[i, j] + 1)

        maxloc = energy.argmax()
        maxloc_row, maxloc_col = np.unravel_index(maxloc, energy.shape)

        Q1_index = CR_50[maxloc_row]
        Q3_index = CR_50[maxloc_col]
        Q1_q = qt[:, Q1_index]
        Q3_q = qt[:, Q3_index]
        Q1 = ft[:, Q1_index]
        Q3 = ft[:, Q3_index]

        # identify amplitude quantiles
        dy_ordering = dy.argsort()
        CR_alpha = dy_ordering[0 : np.round(N * (1 - alpha)).astype("int")]
        tmp = dy[CR_alpha]
        m = tmp.max()
        angle = np.zeros((CR_alpha.shape[0], CR_alpha.shape[0]))
        energy = np.zeros((CR_alpha.shape[0], CR_alpha.shape[0]))
        for i in range(0, CR_alpha.shape[0] - 1):
            for j in range(i + 1, CR_alpha.shape[0]):
                q1 = qt[:, CR_alpha[i]] - q_median
                q3 = qt[:, CR_alpha[j]] - q_median
                q1 /= np.sqrt(trapezoid(q1**2, time))
                q3 /= np.sqrt(trapezoid(q3**2, time))
                angle[i, j] = trapezoid(q1 * q3, time)
                energy[i, j] = (1 - lam) * (
                    dy[CR_alpha[i]] / m + dy[CR_alpha[j]] / m
                ) - lam * (angle[i, j] + 1)

        maxloc = energy.argmax()
        maxloc_row, maxloc_col = np.unravel_index(maxloc, energy.shape)

        Q1a_index = CR_alpha[maxloc_row]
        Q3a_index = CR_alpha[maxloc_col]
        Q1a_q = qt[:, Q1a_index]
        Q3a_q = qt[:, Q3a_index]
        Q1a = ft[:, Q1a_index]
        Q3a = ft[:, Q3a_index]

        # compute amplitude whiskers
        IQR = dy[Q1_index] + dy[Q3_index]
        v1 = Q1_q - q_median
        v3 = Q3_q - q_median
        upper_q = Q3_q + k_a * IQR * v3 / np.sqrt(trapezoid(v3**2, time))
        lower_q = Q1_q + k_a * IQR * v1 / np.sqrt(trapezoid(v1**2, time))

        upper_dis = np.sqrt(trapezoid((upper_q - q_median) ** 2, time))
        lower_dis = np.sqrt(trapezoid((lower_q - q_median) ** 2, time))
        whisker_dis = max(upper_dis, lower_dis)

        # identify amplitude outliers
        outlier_index = np.array([])
        for i in range(0, N):
            if dy[dy_ordering[N - i - 1]] > whisker_dis:
                outlier_index = np.append(outlier_index, dy[dy_ordering[N - i - 1]])

        # identify amplitude extremes
        distance_to_upper = np.full(N, np.inf)
        distance_to_lower = np.full(N, np.inf)
        out_50_CR = np.setdiff1d(np.arange(0, N), outlier_index)
        for i in range(0, out_50_CR.shape[0]):
            j = out_50_CR[i]
            distance_to_upper[j] = np.sqrt(trapezoid((upper_q - qt[:, j]) ** 2, time))
            distance_to_lower[j] = np.sqrt(trapezoid((lower_q - qt[:, j]) ** 2, time))

        max_index = distance_to_upper.argmin()
        min_index = distance_to_lower.argmin()
        min_q = qt[:, min_index]
        max_q = qt[:, max_index]
        minn = ft[:, min_index]
        maxx = ft[:, max_index]

        s = np.linspace(0, 1, 100)
        Fs2 = np.zeros((time.shape[0], 595))
        Fs2[:, 0] = (1 - s[0]) * minn + s[0] * Q1
        for j in range(1, 100):
            Fs2[:, j] = (1 - s[j]) * minn + s[j] * Q1a
            Fs2[:, 98 + j] = (1 - s[j]) * Q1a + s[j] * Q1
            Fs2[:, 197 + j] = (1 - s[j]) * Q1 + s[j] * f_median
            Fs2[:, 296 + j] = (1 - s[j]) * f_median + s[j] * Q3
            Fs2[:, 395 + j] = (1 - s[j]) * Q3 + s[j] * Q3a
            Fs2[:, 494 + j] = (1 - s[j]) * Q3a + s[j] * maxx

        d1 = np.sqrt(trapezoid((q_median - Q1_q) ** 2, time))
        d1a = np.sqrt(trapezoid((Q1_q - Q1a_q) ** 2, time))
        dl = np.sqrt(trapezoid((Q1a_q - min_q) ** 2, time))
        d3 = np.sqrt(trapezoid((q_median - Q3_q) ** 2, time))
        d3a = np.sqrt(trapezoid((Q3_q - Q3a_q) ** 2, time))
        du = np.sqrt(trapezoid((Q3a_q - max_q) ** 2, time))
        part1 = np.linspace(-d1 - d1a - dl, -d1 - d1a, 100)
        part2 = np.linspace(-d1 - d1a, -d1, 100)
        part3 = np.linspace(-d1, 0, 100)
        part4 = np.linspace(0, d3, 100)
        part5 = np.linspace(d3, d3 + d3a, 100)
        part6 = np.linspace(d3 + d3a, d3 + d3a + du, 100)
        allparts = np.hstack(
            (
                part1,
                part2[1:100],
                part3[1:100],
                part4[1:100],
                part5[1:100],
                part6[1:100],
            )
        )
        U, V = np.meshgrid(time, allparts)
        U = np.transpose(U)
        V = np.transpose(V)

        self.Q1 = Q1
        self.Q3 = Q3
        self.Q1a = Q1a
        self.Q3a = Q3a
        self.minn = minn
        self.maxx = maxx
        self.outlier_index = outlier_index
        self.f_median = f_median
        self.q_median = q_median

        plt = collections.namedtuple(
            "plt",
            [
                "U",
                "V",
                "Fs2",
                "allparts",
                "d1",
                "d1a",
                "dl",
                "d3",
                "d3a",
                "du",
                "Q1q",
                "Q3q",
            ],
        )

        self.plt = plt(U, V, Fs2, allparts, d1, d1a, dl, d3, d3a, du, Q1a_q, Q3a_q)
        return

    def plot(self):
        """
        plot box plot and surface plot

        Usage: obj.plot()
        """
        M = self.warp_data.time.shape[0]
        fig1, ax1 = plt.subplots()
        ax1.plot(self.warp_data.time, self.f_median, "k")
        ax1.plot(self.warp_data.time, self.Q1, "b")
        ax1.plot(self.warp_data.time, self.Q3, "b")
        ax1.plot(self.warp_data.time, self.Q1a, "g")
        ax1.plot(self.warp_data.time, self.Q3a, "g")
        ax1.plot(self.warp_data.time, self.minn, "r")
        ax1.plot(self.warp_data.time, self.maxx, "r")

        fig2 = plt.figure()
        ax = fig2.gca(projection="3d")
        ax.plot_surface(
            self.plt.U,
            self.plt.V,
            self.plt.Fs2,
            cmap=cm.viridis,
            rcount=200,
            ccount=200,
        )
        ax.plot(self.warp_data.time, np.zeros(M), self.f_median, "k")
        ax.plot(self.warp_data.time, np.repeat(-self.plt.d1, M), self.Q1, "b")
        ax.plot(
            self.warp_data.time,
            np.repeat(-self.plt.d1 - self.plt.d1a, M),
            self.Q1a,
            "g",
        )
        ax.plot(
            self.warp_data.time,
            np.repeat(-self.plt.d1 - self.plt.d1a - self.plt.dl, M),
            self.minn,
            "r",
        )
        ax.plot(self.warp_data.time, np.repeat(self.plt.d3, M), self.Q3, "b")
        ax.plot(
            self.warp_data.time, np.repeat(self.plt.d3 + self.plt.d3a, M), self.Q3a, "g"
        )
        ax.plot(
            self.warp_data.time,
            np.repeat(self.plt.d3 + self.plt.d3a + self.plt.du, M),
            self.maxx,
            "r",
        )

        plt.show()


class phbox:
    """
    This class provides phase boxplot for functional data using the
    SRVF framework

    Usage:  obj = phbox(warp_data)

    :param warp_data: fdawarp class with alignment data
    :type warp_data: fdawarp
    :param Q1: First quartile
    :param Q3: Second quartile
    :param Q1a: First quantile based on alpha
    :param Q3a: Second quantile based on alpha
    :param minn: minimum extreme function
    :param maxx: maximum extreme function
    :param outlier_index: indexes of outlier functions
    :param median_x: median warping function
    :param psi_median: median srvf of warping function
    :param plt: surface plot mesh

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the phbox class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        if fdawarp.type == "mean":
            raise Exception(
                'Please align fdawarp class using srsf_align with method="median"!'
            )

        self.warp_data = fdawarp

    def construct_boxplot(self, alpha=0.05, k_a=1):
        """
        This function constructs phase boxplot for functional data using the elastic
        square-root slope (srsf) framework.

        :param alpha: quantile value (e.g.,=.05, i.e., 95%)
        :param k_a: scalar for outlier cutoff (e.g.,=1)
        """

        if self.warp_data.rsamps:
            gam = self.warp_data.gams
        else:
            gam = self.warp_data.gam
        M, N = gam.shape
        t = np.linspace(0, 1, M)
        time = t
        lam = 0.5

        # compute phase median
        median_x, psi_median, psi, vec = uf.SqrtMedian(gam)

        # compute phase distances
        dx = np.zeros(N)
        v = np.zeros((M, N))
        for k in range(0, N):
            v[:, k], d = geo.inv_exp_map(psi_median, psi[:, k])
            dx[k] = np.sqrt(trapezoid(v[:, k] ** 2, t))

        dx_ordering = dx.argsort()
        CR_50 = dx_ordering[0 : np.ceil(N / 2).astype("int")]
        tmp = dx[CR_50]
        m = tmp.max()

        # identify phase quartiles
        angle = np.zeros((CR_50.shape[0], CR_50.shape[0]))
        energy = np.zeros((CR_50.shape[0], CR_50.shape[0]))
        for i in range(0, CR_50.shape[0] - 1):
            for j in range(i + 1, CR_50.shape[0]):
                q1 = v[:, CR_50[i]]
                q3 = v[:, CR_50[j]]
                q1 /= np.sqrt(trapezoid(q1**2, time))
                q3 /= np.sqrt(trapezoid(q3**2, time))
                angle[i, j] = trapezoid(q1 * q3, time)
                energy[i, j] = (1 - lam) * (
                    dx[CR_50[i]] / m + dx[CR_50[j]] / m
                ) - lam * (angle[i, j] + 1)

        maxloc = energy.argmax()
        maxloc_row, maxloc_col = np.unravel_index(maxloc, energy.shape)

        Q1_index = CR_50[maxloc_row]
        Q3_index = CR_50[maxloc_col]
        Q1 = gam[:, Q1_index]
        Q3 = gam[:, Q3_index]
        Q1_psi = np.sqrt(np.gradient(Q1, 1 / (M - 1)))
        Q3_psi = np.sqrt(np.gradient(Q3, 1 / (M - 1)))

        # identify phase quantiles
        dx_ordering = dx.argsort()
        CR_alpha = dx_ordering[0 : np.round(N * (1 - alpha)).astype("int")]
        tmp = dx[CR_alpha]
        m = tmp.max()
        angle = np.zeros((CR_alpha.shape[0], CR_alpha.shape[0]))
        energy = np.zeros((CR_alpha.shape[0], CR_alpha.shape[0]))
        for i in range(0, CR_alpha.shape[0] - 1):
            for j in range(i + 1, CR_alpha.shape[0]):
                q1 = v[:, CR_alpha[i]]
                q3 = v[:, CR_alpha[j]]
                q1 /= np.sqrt(trapezoid(q1**2, time))
                q3 /= np.sqrt(trapezoid(q3**2, time))
                angle[i, j] = trapezoid(q1 * q3, time)
                energy[i, j] = (1 - lam) * (
                    dx[CR_alpha[i]] / m + dx[CR_alpha[j]] / m
                ) - lam * (angle[i, j] + 1)

        maxloc = energy.argmax()
        maxloc_row, maxloc_col = np.unravel_index(maxloc, energy.shape)

        Q1a_index = CR_alpha[maxloc_row]
        Q3a_index = CR_alpha[maxloc_col]
        Q1a = gam[:, Q1a_index]
        Q3a = gam[:, Q3a_index]
        Q1a_psi = np.sqrt(np.gradient(Q1a, 1 / (M - 1)))
        Q3a_psi = np.sqrt(np.gradient(Q3a, 1 / (M - 1)))

        # check quartile and quantile going in same direction
        tst = trapezoid(v[:, Q1a_index] * v[:, Q1_index])
        if tst < 0:
            Q1a = gam[:, Q3a_index]
            Q3a = gam[:, Q1a_index]

        # compute phase whiskers
        IQR = dx[Q1_index] + dx[Q3_index]
        v1 = v[:, Q3a_index]
        v3 = v[:, Q3a_index]
        upper_v = v3 + k_a * IQR * v3 / np.sqrt(trapezoid(v3**2, time))
        lower_v = v1 + k_a * IQR * v1 / np.sqrt(trapezoid(v1**2, time))

        upper_dis = np.sqrt(trapezoid(v3**2, time))
        lower_dis = np.sqrt(trapezoid(v1**2, time))
        whisker_dis = max(upper_dis, lower_dis)

        # identify phase outliers
        outlier_index = np.array([])
        for i in range(0, N):
            if dx[dx_ordering[N - 1 - i]] > whisker_dis:
                outlier_index = np.append(outlier_index, dx_ordering[N - i - 1])

        # identify phase extremes
        distance_to_upper = np.full(N, np.inf)
        distance_to_lower = np.full(N, np.inf)
        out_50_CR = np.setdiff1d(np.arange(0, N), outlier_index)
        for i in range(0, out_50_CR.shape[0]):
            j = out_50_CR[i]
            distance_to_upper[j] = np.sqrt(trapezoid((upper_v - v[:, j]) ** 2, time))
            distance_to_lower[j] = np.sqrt(trapezoid((lower_v - v[:, j]) ** 2, time))

        max_index = distance_to_upper.argmin()
        min_index = distance_to_lower.argmin()
        minn = gam[:, min_index]
        maxx = gam[:, max_index]
        min_psi = psi[:, min_index]
        max_psi = psi[:, max_index]

        s = np.linspace(0, 1, 100)
        Fs2 = np.zeros((time.shape[0], 595))
        Fs2[:, 0] = (1 - s[0]) * (minn - t) + s[0] * (Q1 - t)
        for j in range(1, 100):
            Fs2[:, j] = (1 - s[j]) * (minn - t) + s[j] * (Q1a - t)
            Fs2[:, 98 + j] = (1 - s[j]) * (Q1a - t) + s[j] * (Q1 - t)
            Fs2[:, 197 + j] = (1 - s[j]) * (Q1 - t) + s[j] * (median_x - t)
            Fs2[:, 296 + j] = (1 - s[j]) * (median_x - t) + s[j] * (Q3 - t)
            Fs2[:, 395 + j] = (1 - s[j]) * (Q3 - t) + s[j] * (Q3a - t)
            Fs2[:, 494 + j] = (1 - s[j]) * (Q3a - t) + s[j] * (maxx - t)

        d1 = np.sqrt(trapezoid(psi_median * Q1_psi, time))
        d1a = np.sqrt(trapezoid(Q1_psi * Q1a_psi, time))
        dl = np.sqrt(trapezoid(Q1a_psi * min_psi, time))
        d3 = np.sqrt(trapezoid((psi_median * Q3_psi), time))
        d3a = np.sqrt(trapezoid((Q3_psi * Q3a_psi), time))
        du = np.sqrt(trapezoid((Q3a_psi * max_psi), time))
        part1 = np.linspace(-d1 - d1a - dl, -d1 - d1a, 100)
        part2 = np.linspace(-d1 - d1a, -d1, 100)
        part3 = np.linspace(-d1, 0, 100)
        part4 = np.linspace(0, d3, 100)
        part5 = np.linspace(d3, d3 + d3a, 100)
        part6 = np.linspace(d3 + d3a, d3 + d3a + du, 100)
        allparts = np.hstack(
            (
                part1,
                part2[1:100],
                part3[1:100],
                part4[1:100],
                part5[1:100],
                part6[1:100],
            )
        )
        U, V = np.meshgrid(time, allparts)
        U = np.transpose(U)
        V = np.transpose(V)

        self.Q1 = Q1
        self.Q3 = Q3
        self.Q1a = Q1a
        self.Q3a = Q3a
        self.minn = minn
        self.maxx = maxx
        self.outlier_index = outlier_index
        self.median_x = median_x
        self.psi_media = psi_median

        plt = collections.namedtuple(
            "plt",
            [
                "U",
                "V",
                "Fs2",
                "allparts",
                "d1",
                "d1a",
                "dl",
                "d3",
                "d3a",
                "du",
                "Q1_psi",
                "Q3_psi",
            ],
        )

        self.plt = plt(U, V, Fs2, allparts, d1, d1a, dl, d3, d3a, du, Q1a_psi, Q3a_psi)
        return

    def plot(self):
        """
        plot box plot and surface plot

        Usage: obj.plot()
        """

        M = self.warp_data.time.shape[0]
        time = np.linspace(0, 1, M)
        fig1, ax1 = plt.subplots()
        ax1.plot(time, self.median_x, "k")
        ax1.plot(time, self.Q1, "b")
        ax1.plot(time, self.Q3, "b")
        ax1.plot(time, self.Q1a, "g")
        ax1.plot(time, self.Q3a, "g")
        ax1.plot(time, self.minn, "r")
        ax1.plot(time, self.maxx, "r")
        ax1.set_aspect("equal")

        fig2 = plt.figure()
        ax = fig2.gca(projection="3d")
        ax.plot_surface(
            self.plt.U,
            self.plt.V,
            self.plt.Fs2,
            cmap=cm.viridis,
            rcount=200,
            ccount=200,
        )
        ax.plot(time, np.zeros(M), self.median_x - time, "k")
        ax.plot(time, np.repeat(-self.plt.d1, M), self.Q1 - time, "b")
        ax.plot(time, np.repeat(-self.plt.d1 - self.plt.d1a, M), self.Q1a - time, "g")
        ax.plot(
            time,
            np.repeat(-self.plt.d1 - self.plt.d1a - self.plt.dl, M),
            self.minn - time,
            "r",
        )
        ax.plot(time, np.repeat(self.plt.d3, M), self.Q3 - time, "b")
        ax.plot(time, np.repeat(self.plt.d3 + self.plt.d3a, M), self.Q3a - time, "g")
        ax.plot(
            time,
            np.repeat(self.plt.d3 + self.plt.d3a + self.plt.du, M),
            self.maxx - time,
            "r",
        )

        plt.show()

        return



import numpy as np


def kernel(x, y, l2):
    # RBF Kernel
    sqdist = np.sum(x**2, 1).reshape(-1, 1) + np.sum(y**2, 1) - 2 * np.dot(x, y.T)
    return np.exp(-0.5 * (1 / l2) * sqdist)


def gp_posterior(X, y, Xtest, l2=0.1, noise_var=1e-6):
    # compute the mean at our test points.
    N, n = len(X), len(Xtest)
    K = kernel(X, X, l2)
    L = np.linalg.cholesky(K + noise_var * np.eye(N))
    Lk = np.linalg.solve(L, kernel(X, Xtest, l2))
    mu = np.dot(Lk.T, np.linalg.solve(L, y))
    # compute the variance at our test points.
    K_ = kernel(Xtest, Xtest, l2)
    sd = np.sqrt(np.diag(K_) - np.sum(Lk**2, axis=0))
    return (mu, sd)
